/*
*  Copyright 2016, University of Texas at Austin, W. Cyrus Proctor
*  Author: W. Cyrus Proctor <cproctor@tacc.utexas.edu>
*
*    This file is part of OpenMFA.
* 
*    OpenMFA is free software: you can redistribute it and/or
*    modify it under the terms of the GNU Affero General Public
*    License as published by the Free Software Foundation, either
*    version 3 of the License, or (at your option) any later
*    version.
* 
*    OpenMFA is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty
*    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
*    the GNU Affero General Public License for more details.
* 
*    You should have received a copy of the GNU Affero General
*    Public License along with OpenMFA. If not, see
*    <http://www.gnu.org/licenses/>.
*/

#ifdef sun
#else
#define _XOPEN_SOURCE
#include <time.h>
#endif


#define PAM_SM_AUTH

#include <stdlib.h>
#include "pam_openmfa.h"

//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------


//-----------------------------------------------------------------------------------
// Command-line options for module
//-----------------------------------------------------------------------------------

struct options_t
{
  int debug;
  int read_delay;
  int time_diff;
};
typedef struct options_t options_t;

//-----------------------------------------------------------------------------------
// Parse command-line options for module 
//-----------------------------------------------------------------------------------

static void
parse_option(pam_handle_t *pamh, const char *argv, options_t *options)
{
  if(strcasecmp (argv, "debug") == 0)
  {
    options->debug = 1;
    openmfa_log(pamh,OPENMFA_INFO,"openmfa:parse_option: debug mode activated '%s'", argv);
  }
  else if(strncasecmp (argv, "read_delay=", 11) == 0)
  {
    options->read_delay = strtol(&argv[11], NULL, 10);
    if(options->read_delay < 0)
    {
      options->read_delay = 0;
    }
    if(options->read_delay > 5)
    {
      options->read_delay = 5;
    }
    openmfa_log(pamh,OPENMFA_INFO,"openmfa:parse_option: read_delay set to '%d'", options->read_delay);
  }
  else if(strncasecmp (argv, "time_diff=", 10) == 0)
  {
    options->time_diff = strtol(&argv[10], NULL, 10);
    if(options->time_diff < 0)
    {
      options->time_diff = 0;
    }
    if(options->time_diff > 300)
    {
      options->time_diff = 300;
    }
    openmfa_log(pamh,OPENMFA_INFO,"openmfa:parse_option: time_diff set to '%d'", options->time_diff);
  }
  else
  {
    openmfa_log(pamh,OPENMFA_ERR,"openmfa:parse_option: unknown option '%s'", argv);
  }
}

//-----------------------------------------------------------------------------------
// Pam authentication function
//-----------------------------------------------------------------------------------
PAM_EXTERN
int 
pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc, const char **argv) 
{

  options_t options;
  memset (&options, 0, sizeof (options));
  // Default values
  options.debug        = 0;  // debug turned off
  options.read_delay   = 1;  // time delay before reading /var/log/secure (s)
  options.time_diff    = 30; // time difference between pubkey success and now (s)

  //-----------------------------------------------------------------------------------
  // Parse parameters for module
  //-----------------------------------------------------------------------------------

  for ( ; argc-- > 0; argv++)
  {
    parse_option(pamh, *argv, &options);
  }

  //-----------------------------------------------------------------------------------
  // Get local user name
  //-----------------------------------------------------------------------------------

  int pgu_rc;
#ifdef sun
  char *user = NULL;
#else
  const char *user = NULL;
#endif
  static char username_prompt[] = "OPENMFA Username:";
  pgu_rc = pam_get_user(pamh, &user, username_prompt);
  if(pgu_rc == PAM_SUCCESS && user != NULL)
  {
    if(options.debug)
      openmfa_log(pamh,OPENMFA_INFO,"openmfa:get_user_name: acquired username %s", user);
  }
  else
  { 
    openmfa_log(pamh,OPENMFA_ERR,"openmfa:get_user_name:FAIL: cannot acquire username; rc: %d", pgu_rc);
    return(PAM_SYSTEM_ERR);
  }

  //-----------------------------------------------------------------------------------
  // Get remote host
  //-----------------------------------------------------------------------------------
  
  int pgh_rc;
#ifdef sun
  void *rhost = NULL;
#else
  const void *rhost = NULL;
#endif
  pgh_rc = pam_get_item(pamh, PAM_RHOST, &rhost);
  if(pgh_rc == PAM_SUCCESS && rhost != NULL)
  {
    if(options.debug)
      openmfa_log(pamh,OPENMFA_INFO,"openmfa:get_remote_host: acquired hostname %s", rhost);
  }
  else
  { 
    openmfa_log(pamh,OPENMFA_ERR,"openmfa:get_remote_host:FAIL: cannot acquire remote host IP; rc: %d", pgh_rc);
    return(PAM_SYSTEM_ERR);
  }
  
  //-----------------------------------------------------------------------------------
  // Delay before reading /var/log/secure
  //-----------------------------------------------------------------------------------
  
  sleep(options.read_delay);

  //-----------------------------------------------------------------------------------
  // Query /var/log/secure for partial publickey success
  //-----------------------------------------------------------------------------------

  FILE *vls;
  char command[512];
  char log_msg[512];
  bool pubkey_success = false;

  // Set the command to find first reverse occurance of string in /var/log/secure
  sprintf(command,"tac /var/log/secure | grep \"Partial publickey for %s from %s\" | head -n 1",user,rhost);
  //Open the command for reading.
  vls = popen(command, "r");
  if(vls == NULL) 
  {
    openmfa_log(pamh,OPENMFA_ERR,"openmfa:get_read_log_secure:FAIL: cannot read /var/log/secure");
  }
  if(fgets(log_msg, sizeof(log_msg)-1, vls) == NULL) 
  {
    if(options.debug)
      openmfa_log(pamh, OPENMFA_INFO,"openmfa:get_read_log_secure: pubkey success not found");
  }
  else
  {
    // Get current time
    time_t rawtime;
    time_t current_epoch_time;
    struct tm *current_tm;
    time(&rawtime);
    current_tm = localtime(&rawtime);
    current_epoch_time = mktime(current_tm);
    
    // Get log time
    time_t log_epoch_time;
    struct tm log_tm;
    memset(&log_tm, 0, sizeof(struct tm));
    char log_msg_p_year[522];
    // Copy in the current year b/c log doesn't display this
    sprintf(log_msg_p_year,"%d %s",current_tm->tm_year+1900,log_msg);
    if (strptime(log_msg_p_year, "%Y %b %d %H:%M:%S", &log_tm) == NULL)
    {
      openmfa_log(pamh,OPENMFA_ERR,"openmfa:get_read_log_secure:FAIL: Cannot get log time.");
    }
    log_tm.tm_isdst = -1;
    log_epoch_time = mktime(&log_tm);
    time_t time_diff;
    time_diff = current_epoch_time - log_epoch_time;
    if(time_diff >= 0 && time_diff <= options.time_diff)
    {
      pubkey_success = true;
      if(options.debug)
        openmfa_log(pamh, OPENMFA_INFO,"openmfa:get_read_log_secure: pubkey check success");
    }
    else
    {
      pubkey_success = false;
      if(options.debug)
        openmfa_log(pamh, OPENMFA_INFO,"openmfa:get_read_log_secure: pubkey check failure");
    }
    if(options.debug)
    {
      openmfa_log(pamh, OPENMFA_INFO,"openmfa:get_read_log_secure:   Year: %d", log_tm.tm_year);
      openmfa_log(pamh, OPENMFA_INFO,"openmfa:get_read_log_secure:  Month: %d", log_tm.tm_mon);
      openmfa_log(pamh, OPENMFA_INFO,"openmfa:get_read_log_secure:    Day: %d", log_tm.tm_mday);
      openmfa_log(pamh, OPENMFA_INFO,"openmfa:get_read_log_secure:   Hour: %d", log_tm.tm_hour);
      openmfa_log(pamh, OPENMFA_INFO,"openmfa:get_read_log_secure: Minute: %d", log_tm.tm_min);
      openmfa_log(pamh, OPENMFA_INFO,"openmfa:get_read_log_secure: Second: %d", log_tm.tm_sec);
      openmfa_log(pamh, OPENMFA_INFO,"openmfa:get_read_log_secure: Is DST: %d", log_tm.tm_isdst);
      openmfa_log(pamh, OPENMFA_INFO,"openmfa:get_read_log_secure: CEpoch: %d", current_epoch_time);
      openmfa_log(pamh, OPENMFA_INFO,"openmfa:get_read_log_secure: LEpoch: %d", log_epoch_time);
      openmfa_log(pamh, OPENMFA_INFO,"openmfa:get_read_log_secure:  TDiff: %d", time_diff);
    }

  }
  pclose(vls);
 
  if(pubkey_success)
  {
    return(PAM_SUCCESS);
  }

  return(PAM_AUTH_ERR);
}

//-----------------------------------------------------------------------------------
// Pam set credentials function -- not handled in this module
//-----------------------------------------------------------------------------------

PAM_EXTERN int
pam_sm_setcred(pam_handle_t *pamh, int flags,
int argc, const char *argv[])
{
      (void)pamh;
      (void)flags;
      (void)argc;
      (void)argv;
      return (PAM_SUCCESS);
}
