/*
*  Copyright 2016, University of Texas at Austin, W. Cyrus Proctor
*  Author: W. Cyrus Proctor <cproctor@tacc.utexas.edu>
*
*    This file is part of OpenMFA.
* 
*    OpenMFA is free software: you can redistribute it and/or
*    modify it under the terms of the GNU Affero General Public
*    License as published by the Free Software Foundation, either
*    version 3 of the License, or (at your option) any later
*    version.
* 
*    OpenMFA is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty
*    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
*    the GNU Affero General Public License for more details.
* 
*    You should have received a copy of the GNU Affero General
*    Public License along with OpenMFA. If not, see
*    <http://www.gnu.org/licenses/>.
*/


#define PAM_SM_AUTH

#ifdef sun
#define _POSIX_PTHREAD_SEMANTICS
#endif

#include <stdio.h>
#include <pwd.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#include "pam_openmfa_token.h"
#include "pam_openmfa.h"
#include "radius.h"




//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------

enum enforcement
{
  OFF,
  PAIRED,
  COUNTDOWN,
  FULL
};

enum hostlist
{
  NONE,
  DEFAULT
};

//-----------------------------------------------------------------------------------
// Command-line options for module
//-----------------------------------------------------------------------------------

struct options_t
{
  int debug;
  int enforce;
  int hostlist;
  bool date;
  long countdown;
  bool is_website;
  const char *website;
};
typedef struct options_t options_t;

//-----------------------------------------------------------------------------------
// Build Radius Request
//-----------------------------------------------------------------------------------

static int
build_radius_request(pam_handle_t *pamh, struct rad_handle **radh, const char *user, const char *token_code, struct options_t *options)
{
  if( ! (*radh = rad_auth_open()) )
  {
    openmfa_log(pamh,OPENMFA_ERR,"openmfa:build_radius_request: rad_auth_open returned null radh");
    return(PAM_SYSTEM_ERR);
  }
  int rad_port      = RAD_SRV_PORT;
  int rad_timeout   = 1;
  int rad_max_tries = 1;

  if( options->hostlist == DEFAULT )
  {
    const wchar_t rad_host_a [] = RAD_SRV_HOST_A;
    const wchar_t rad_host_b [] = RAD_SRV_HOST_B;
    const wchar_t rad_key_a  [] = RAD_SRV_KEY_A;
    const wchar_t rad_key_b  [] = RAD_SRV_KEY_B;
    int i;
    char rad_secret1   [SLEN];

    for ( i = 0; i < SLEN; i++)
    {
      rad_secret1[i] = rad_key_a[i] ^ rad_key_b[i];
    }
    char rad_hostname1 [HLEN];
    for ( i = 0; i < HLEN; i++)
    {
      rad_hostname1[i] = rad_host_a[i] ^ rad_host_b[i];
    }
     
    int ras_rc;
    ras_rc = rad_add_server(*radh, rad_hostname1, rad_port, rad_secret1, rad_timeout, rad_max_tries);
    if(ras_rc < 0)
    {
      openmfa_log(pamh,OPENMFA_ERR,"openmfa:build_radius_request: failed to add server: %s\n", rad_strerror(*radh));
      return(PAM_SYSTEM_ERR);
    }
  }
  else
  {
    openmfa_log(pamh,OPENMFA_ERR,"openmfa:build_radius_request: unexpected hostlist option: %s\n", options->hostlist);
    return(PAM_SYSTEM_ERR);
  }

  int rcs_rc;
  rcs_rc = rad_create_request(*radh, RAD_ACCESS_REQUEST);
  if(rcs_rc < 0)
  {
    openmfa_log(pamh,OPENMFA_ERR,"openmfa:build_radius_request: failed to create request: %s\n", rad_strerror(*radh));
    return(PAM_SYSTEM_ERR);
  }
  
  rad_put_string(*radh, RAD_USER_NAME, user);
  rad_put_string(*radh, RAD_USER_PASSWORD, token_code);

  return 100; //success
}

//-----------------------------------------------------------------------------------
// Parse command-line options for module 
//-----------------------------------------------------------------------------------

static void
parse_option(pam_handle_t *pamh, const char *argv, options_t *options)
{
  if(strcasecmp (argv, "debug") == 0)
  {
    options->debug = 1;
    openmfa_log(pamh,OPENMFA_INFO,"openmfa:parse_option: debug mode activated '%s'", argv);
  }
  else if(strncasecmp (argv, "enforce=", 8) == 0)
  {
    if(strncasecmp (&argv[8], "full", 4) == 0)
    {
      options->enforce = FULL;
      openmfa_log(pamh,OPENMFA_INFO,"openmfa:parse_option: enforce=full activated '%s'", argv);
    }
    else if(strncasecmp (&argv[8], "countdown", 9) == 0)
    {
      options->enforce = COUNTDOWN;
      openmfa_log(pamh,OPENMFA_INFO,"openmfa:parse_option: enforce=countdown activated '%s'", argv);
    }
    else if(strncasecmp (&argv[8], "paired", 6) == 0)
    {
      options->enforce = PAIRED;
      openmfa_log(pamh,OPENMFA_INFO,"openmfa:parse_option: enforce=paired activated '%s'", argv);
    }
    else if(strncasecmp (&argv[8], "off", 3) == 0)
    {
      options->enforce = OFF;
      openmfa_log(pamh,OPENMFA_INFO,"openmfa:parse_option: enforce=off activated '%s'", argv);
    }
    else
    {
      openmfa_log(pamh,OPENMFA_ERR,"openmfa:parse_option: enforce= unknown option '%s'; ignoring", argv);
    }
  }
  else if(strncasecmp (argv, "date=", 5) == 0)
  {
    // Get current time
    time_t rawtime;
    time_t current_epoch_time;
    struct tm *current_tm;
    time(&rawtime);
    current_tm = localtime(&rawtime);
    current_epoch_time = mktime(current_tm);
    // Read in countdown time
    struct tm countdown_tm;
    countdown_tm = *current_tm; // Set them equal to begin with
    if ( ! strptime(&argv[5], "%Y-%m-%d", &countdown_tm) )
    {
      openmfa_log(pamh,OPENMFA_ERR,"openmfa:parse_option:FAIL: Cannot read countdown date.");
      options->enforce = FULL; // Fail to full enforcement
    }
    time_t countdown_epoch_time;
    countdown_epoch_time = mktime(&countdown_tm);
    time_t time_diff;
    time_diff = countdown_epoch_time - current_epoch_time;
    options->countdown = time_diff / 86400.0;
    if(options->debug)
    {
      openmfa_log(pamh,OPENMFA_INFO,"openmfa:parse_option: current epoch time %d", current_epoch_time);
      openmfa_log(pamh,OPENMFA_INFO,"openmfa:parse_option: countdown epoch time %d", countdown_epoch_time);
      openmfa_log(pamh,OPENMFA_INFO,"openmfa:parse_option: time diff %d", time_diff);
    }
    openmfa_log(pamh,OPENMFA_INFO,"openmfa:parse_option: countdown date delta in days %d", options->countdown);
    options->date = true; // successful date parse
  } 
  else if(strncasecmp (argv, "website=", 8) == 0)
  {
    options->website = &argv[8];
    openmfa_log(pamh,OPENMFA_ERR,"openmfa:parse_option: website set to '%s'", options->website);
    options->is_website = true;
  }
  else if(strncasecmp (argv, "hostlist=", 9) == 0)
  {
    if(strncasecmp (&argv[9], "default", 6) == 0)
    {
      options->hostlist = DEFAULT;
      openmfa_log(pamh,OPENMFA_INFO,"openmfa:parse_option: hostlist=default activated '%s'", argv);
    }
    else
    {
      options->hostlist = NONE;
      openmfa_log(pamh,OPENMFA_ERR,"openmfa:parse_option: hostlist= unknown option '%s'; ignoring", argv);
    }
  }
  else
  {
    openmfa_log(pamh,OPENMFA_ERR,"openmfa:parse_option: unknown option '%s'", argv);
  }
}


//-----------------------------------------------------------------------------------
// Pam authentication function
//-----------------------------------------------------------------------------------

PAM_EXTERN
int 
pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc, const char **argv) 
{
  
  options_t options;
  memset (&options, 0, sizeof (options));
  // Default values
  options.debug        = 0;       // debug turned off
  options.enforce      = FULL;    // full enforcement turned on
  options.hostlist     = DEFAULT; // use default hostlist
  options.date         = false;   // no countdown date present
  options.countdown    = 0;       // no countdown
  options.is_website   = false;   // no website url
  options.website      = "N/A";   // no website url

  //-----------------------------------------------------------------------------------
  // Parse parameters for module
  //-----------------------------------------------------------------------------------

  for ( ; argc-- > 0; argv++)
  {
    parse_option(pamh, *argv, &options);
  }

  if( options.enforce == COUNTDOWN && !( options.date && options.is_website ) )
  {
    openmfa_log(pamh,OPENMFA_ERR,"openmfa:parse_option: ERROR: date and website option must be specified with countdown enforcement");
    openmfa_log(pamh,OPENMFA_ERR,"openmfa:parse_option: ERROR: failing to full enforcement");
    options.enforce = FULL;
  }
  else if(options.enforce == COUNTDOWN && options.date && options.is_website && (options.countdown < 0))
  {
    openmfa_log(pamh,OPENMFA_INFO,"openmfa:parse_option: countdown date has passed; full enforcement");
    options.enforce = FULL;
  }
  else if(options.enforce != COUNTDOWN && (options.date || options.is_website) )
  {
    openmfa_log(pamh,OPENMFA_ERR,"openmfa:parse_option: ERROR: date and website options must ONLY be specified with countdown enforcement");
    openmfa_log(pamh,OPENMFA_ERR,"openmfa:parse_option: ERROR: failing to full enforcement");
    options.enforce = FULL;
  }
  else if(options.enforce == OFF)
  {
    openmfa_log(pamh,OPENMFA_INFO,"openmfa:parse_option: enforcement is set to off; exiting with pam success");
    return(PAM_SUCCESS);
  }

  //-----------------------------------------------------------------------------------
  // Get local user name
  //-----------------------------------------------------------------------------------

  int pgu_rc;
#ifdef sun
  char *user = NULL;
#else
  const char *user = NULL;
#endif
  static char username_prompt[] = "OPENMFA Username:";
  pgu_rc = pam_get_user(pamh, &user, username_prompt);
  if(pgu_rc == PAM_SUCCESS && user != NULL)
  {
    if(options.debug)
      openmfa_log(pamh,OPENMFA_INFO,"openmfa:get_user_name: acquired username %s", user);
  }
  else
  { 
    openmfa_log(pamh,OPENMFA_ERR,"openmfa:get_user_name:FAIL: cannot acquire username; rc: %d", pgu_rc);
    return(PAM_SYSTEM_ERR);
  }

  //-----------------------------------------------------------------------------------
  // Get remote host
  //-----------------------------------------------------------------------------------
  
  int pgh_rc;
#ifdef sun
  void *rhost = NULL;
#else
  const void *rhost = NULL;
#endif
  pgh_rc = pam_get_item(pamh, PAM_RHOST, &rhost);
  if(pgh_rc == PAM_SUCCESS && rhost != NULL)
  {
    if(options.debug)
      openmfa_log(pamh,OPENMFA_INFO,"openmfa:get_remote_host: acquired hostname %s", rhost);
  }
  else
  { 
    openmfa_log(pamh,OPENMFA_ERR,"openmfa:get_remote_host:FAIL: cannot acquire remote host IP; rc: %d", pgh_rc);
    return(PAM_SYSTEM_ERR);
  }

  //-----------------------------------------------------------------------------------
  // Get password struct
  //-----------------------------------------------------------------------------------

  // First, get max length of buffer for password struct
  long int init_len = sysconf(_SC_GETPW_R_SIZE_MAX);
  size_t len;
  if(init_len == -1)
  {
    // Default initial length
    len = 1024;
  }
  else
  {
    len = (size_t) init_len;
  }
  char *buffer = (char*)malloc(len);
  if(buffer == NULL)
  {
    openmfa_log(pamh,OPENMFA_ERR,"openmfa:get_password_struct:FAIL: cannot set buffer of length %d",len);
    return(PAM_SYSTEM_ERR);
  }
  int gps_rc;
  struct passwd ps;
  struct passwd *ps_result;
  gps_rc = getpwnam_r(user, &ps, buffer, len, &ps_result);
  if(gps_rc == 0 && ps_result != NULL && ps_result->pw_dir != NULL && ps_result->pw_dir[0] == '/')
  {
    if(options.debug)
      openmfa_log(pamh,OPENMFA_INFO,"openmfa:get_password_struct: pw struct acquired; rc: %d",gps_rc);
  }
  else
  {
    if(options.debug)
      openmfa_log(pamh,OPENMFA_INFO,"openmfa:get_password_struct: cannot access pw struct; rc: %d",gps_rc);
  }
// todo: check usage before free
  free(buffer); buffer = NULL;


  enum auth_types
  {
    OPENMFA,
    OPENMFA_SMS,
    OPENMFA_COUNTDOWN
  };
  int default_auth_type = OPENMFA;
  int auth_type;
  auth_type = default_auth_type;
  FILE *ltqh, *ltqs, *ltqsms;
  char commandh[512], commands[512], commandsms[512];
  char lresh[512], lress[512], lressms[512];
  bool lh = false, ls = false, lsms = false; 

  // Set the commands to query LDAP server for token authentication method (soft, hard, or sms)
#ifdef sun
  sprintf(commandh  ,"ldapsearch -h ldap.foo.com -b \"dc=foo,dc=com\"  \"(uid=%s)\" | grep \"authorizedService:\" | grep \"openmfa-hard-token\"",user);
  sprintf(commands  ,"ldapsearch -h ldap.foo.com -b \"dc=foo,dc=com\"  \"(uid=%s)\" | grep \"authorizedService:\" | grep \"openmfa-soft-token\"",user);
  sprintf(commandsms,"ldapsearch -h ldap.foo.com -b \"dc=foo,dc=com\"  \"(uid=%s)\" | grep \"authorizedService:\" | grep \"openmfa-sms-token\"" ,user);
#else
  sprintf(commandh  ,"ldapsearch -x -H ldap://ldap.foo.com -b \"dc=foo,dc=com\"  \"(uid=%s)\" | grep \"authorizedService:\" | grep \"openmfa-hard-token\"",user);
  sprintf(commands  ,"ldapsearch -x -H ldap://ldap.foo.com -b \"dc=foo,dc=com\"  \"(uid=%s)\" | grep \"authorizedService:\" | grep \"openmfa-soft-token\"",user);
  sprintf(commandsms,"ldapsearch -x -H ldap://ldap.foo.com -b \"dc=foo,dc=com\"  \"(uid=%s)\" | grep \"authorizedService:\" | grep \"openmfa-sms-token\"" ,user);
#endif
  //Open the command for reading
  ltqh   = popen(commandh  , "r");
  ltqs   = popen(commands  , "r");
  ltqsms = popen(commandsms, "r");
  if(ltqh == NULL && ltqs == NULL && ltqsms == NULL) 
  {
    openmfa_log(pamh,OPENMFA_ERR,"openmfa:ldap_token_query: no authorized service found from ldap");
  }
  if(fgets(lresh, sizeof(lresh)-1, ltqh) == NULL) 
  {
    if(options.debug)
      openmfa_log(pamh, OPENMFA_INFO,"openmfa:ldap_token_query: authorized service openmfa-hard-token not found");
  }
  else
  {
    openmfa_log(pamh, OPENMFA_INFO,"openmfa:ldap_token_query: openmfa-hard-token result %s",lresh);
    lh = true;
  }
  if(fgets(lress, sizeof(lress)-1, ltqs) == NULL) 
  {
    if(options.debug)
      openmfa_log(pamh, OPENMFA_INFO,"openmfa:ldap_token_query: authorized service openmfa-soft-token not found");
  }
  else
  {
    openmfa_log(pamh, OPENMFA_INFO,"openmfa:ldap_token_query: openmfa-soft-token result %s",lress);
    ls = true;
  }
  if(fgets(lressms, sizeof(lressms)-1, ltqsms) == NULL) 
  {
    if(options.debug)
      openmfa_log(pamh, OPENMFA_INFO,"openmfa:ldap_token_query: authorized service openmfa-sms-token not found");
  }
  else
  {
    openmfa_log(pamh, OPENMFA_INFO,"openmfa:ldap_token_query: openmfa-sms-token result %s",lressms);
    lsms = true;
  }
  pclose(ltqh);
  pclose(ltqs);
  pclose(ltqsms);

  // No valid ldap authorization services found
  if( options.enforce != COUNTDOWN && (!lh) && (!ls) && (!lsms) )
  {
    openmfa_log(pamh,OPENMFA_INFO,"openmfa:ldap_token_query: all ldap authorized services are false");
    if( options.enforce == FULL )
    {
      openmfa_log(pamh,OPENMFA_ERR,"openmfa:ldap_token_query: enforcement set to full; no services found; failure imminent; continuing anyway");
    }
    else
    {
      openmfa_log(pamh,OPENMFA_INFO,"openmfa:ldap_token_query: enforcement set to paired; no services found; exiting with success");
      return (PAM_SUCCESS);
    }
  }

  // Take preference to SMS if paired
  if(lsms)
  {
    auth_type = OPENMFA_SMS;
    if(options.debug)
      openmfa_log(pamh,OPENMFA_INFO,"openmfa:ldap_token_query: setting auth_type to OPENMFA_SMS");
  }
  else if(ls || lh)
  {
    auth_type = OPENMFA;
    if(options.debug)
      openmfa_log(pamh,OPENMFA_INFO,"openmfa:ldap_token_query: setting auth_type to OPENMFA");
  }

  // If no pairings found and in countdown enforcement mode
  if( options.enforce == COUNTDOWN && (!lsms) && (!ls) && (!lh))
  {
    auth_type = OPENMFA_COUNTDOWN;
    if(options.debug)
      openmfa_log(pamh,OPENMFA_INFO,"openmfa:ldap_token_query: setting auth_type to OPENMFA_COUNTDOWN");
  }


  bool sms_submitted         = false;
  bool sms_already_submitted = false;
  if(auth_type == OPENMFA_SMS)
  {

    //-----------------------------------------------------------------------------------
    // Build Null Radius Request
    //-----------------------------------------------------------------------------------
 
    struct rad_handle *radh_sms = NULL;
    int bnr_rc;
    bnr_rc = build_radius_request(pamh, &radh_sms, user, "", &options);
    if(bnr_rc == 100 && radh_sms != NULL)
    {
      if(options.debug)
        openmfa_log(pamh,OPENMFA_INFO,"openmfa:build_null_request: was successful");
    }
    else
    {
      openmfa_log(pamh,OPENMFA_ERR,"openmfa:build_null_request: system error %d", bnr_rc);
      return(PAM_SYSTEM_ERR);
    }

    //-----------------------------------------------------------------------------------
    // Send Null Radius Request
    //-----------------------------------------------------------------------------------

    int rnr_rc;
    rnr_rc = rad_send_request(radh_sms);
    if(rnr_rc == RAD_ACCESS_ACCEPT)
    {
      if(options.debug)
        openmfa_log(pamh,OPENMFA_INFO,"openmfa:send_radius_null_request: accepted request");
    }
    else if (rnr_rc == RAD_ACCESS_CHALLENGE)
    {
      if(options.debug)
        openmfa_log(pamh,OPENMFA_INFO,"openmfa:send_radius_null_request: challenge request");
    }
    else
    {
      if(options.debug)
        openmfa_log(pamh,OPENMFA_INFO,"openmfa:send_radius_null_request: rejected request: %d %s\n", rnr_rc, rad_strerror(radh_sms));
    }
    
    int att_num;
    const void *att_val;
    size_t att_len;
    char *s;
    while ((att_num = rad_get_attr(radh_sms, &att_val, &att_len)) > 0) 
    {
      s = rad_cvt_string(att_val, att_len);
      if(options.debug)
        openmfa_log(pamh,OPENMFA_INFO,"openmfa: att_num %d att_len %d att_val %s str %s\n", att_num, att_len, att_val, s);
      if(strstr(s, "sms submitted") != NULL)
      {
        sms_submitted = true;
      }
      else if(strstr(s, "already submitted") != NULL)
      {
        sms_already_submitted = true;
      }
      free(s); s = NULL;
    }
  }

  //-----------------------------------------------------------------------------------
  // Prompt for the Token Code
  //-----------------------------------------------------------------------------------

  struct pam_conv *conv;
  struct pam_message msg;
  struct pam_response *response;
  char *token_code;
  int gpc_rc;
#ifdef sun
  struct pam_message *msgp;
  gpc_rc = pam_get_item(pamh, PAM_CONV, (void **)&conv);
#else
  const struct pam_message *msgp;
  gpc_rc = pam_get_item(pamh, PAM_CONV, (const void **)&conv);
#endif
  if(gpc_rc != PAM_SUCCESS)
  {
    openmfa_log(pamh,OPENMFA_ERR,"openmfa:prompt_token_code: pam_get_item non-zero exit: %d", gpc_rc);
    return(PAM_SYSTEM_ERR);
  }
  msg.msg_style = PAM_PROMPT_ECHO_OFF;
  if(sms_submitted)
  {
    msg.msg = "SMS Submitted\nOPENMFA Token Code:";
  }
  else if(sms_already_submitted)
  {
    msg.msg = "SMS Already Submitted\nOPENMFA Token Code:";
  }
  else
  {
    if(auth_type == OPENMFA_COUNTDOWN)
    {
      char countdown_msg[1024];
      snprintf(countdown_msg, sizeof(countdown_msg), 
      "\n---------------------------------------------------------------------\
       \n--------------------OPENMFA Token Authentication---------------------\
       \n---------------------------------------------------------------------\
       \nNo paired devices found.\
       \n%ld days left to register a device for multi-factor authentication.\
       \nPlease visit:\
       \n\
       \n  %s\
       \n\
       \nfor more information.\
       \n---------------------------------------------------------------------\
       \n---------------------------------------------------------------------\
       \n---------------------------------------------------------------------\
       \n\
       \nPress 'Enter' to acknowledge:", options.countdown, options.website);
      msg.msg = countdown_msg;
    }
    else
    {
      msg.msg = "OPENMFA Token Code:";
    }
  }
  msgp = &msg;
  response = NULL;
  int ptc_rc;
  ptc_rc = (*conv->conv)(1, &msgp, &response, conv->appdata_ptr);
  if(ptc_rc == PAM_SUCCESS)
  {
    if(response != NULL)
    {
      token_code = response->resp;
      free(response); response = NULL;
      if(options.debug)
        openmfa_log(pamh,OPENMFA_INFO,"openmfa:prompt_token_code: user %s input token code: '%s'", user, token_code);
    }
    else
    {
      free(response); response = NULL;
      free(token_code); token_code = NULL;
      openmfa_log(pamh,OPENMFA_ERR,"openmfa:prompt_token_code: pam_conv non-zero exit: %d", ptc_rc);
      return(PAM_SYSTEM_ERR);
    }
  }
  else
  {
    free(response); response = NULL;
    free(token_code); token_code = NULL;
    openmfa_log(pamh,OPENMFA_ERR,"openmfa:prompt_token_code: pam_conv non-zero exit: %d", ptc_rc);
    return(PAM_AUTH_ERR);
  }
   
  // Exit out early if in countdown mode 
  if(auth_type == OPENMFA_COUNTDOWN)
  {
    openmfa_log(pamh,OPENMFA_INFO,"openmfa:prompt_token_code: user %s no pairings; countdown mode; exiting with pam success", user);
    return(PAM_SUCCESS);
  }

  //-----------------------------------------------------------------------------------
  // Build Radius Request
  //-----------------------------------------------------------------------------------

  struct rad_handle *radh = NULL;
  int brr_rc;
  brr_rc = build_radius_request(pamh, &radh, user, token_code, &options);
  if(brr_rc == 100 && radh != NULL)
  {
    if(options.debug)
      openmfa_log(pamh,OPENMFA_INFO,"openmfa:build_radius_request: was successful");
  }
  else
  {
    openmfa_log(pamh,OPENMFA_INFO,"openmfa:build_radius_request: system error %d", brr_rc);
    return(PAM_SYSTEM_ERR);
  }

  //-----------------------------------------------------------------------------------
  // Send Radius Request
  //-----------------------------------------------------------------------------------

  int rsr_rc;
  rsr_rc = rad_send_request(radh);
  free(token_code); token_code = NULL;
  if(rsr_rc == RAD_ACCESS_ACCEPT)
  {
    openmfa_log(pamh,OPENMFA_INFO,"openmfa:send_radius_request: accepted request %d %s\n", rsr_rc, rad_strerror(radh));
    return(PAM_SUCCESS);
  }
  else
  {
    openmfa_log(pamh,OPENMFA_INFO,"openmfa:send_radius_request: rejected request: %d %s\n", rsr_rc, rad_strerror(radh));
    return(PAM_AUTH_ERR);
  }  

  // Should never get here
  return(PAM_SYSTEM_ERR);
//  return(PAM_SUCCESS);
}


//-----------------------------------------------------------------------------------
// Pam set credentials function -- not handled in this module
//-----------------------------------------------------------------------------------

PAM_EXTERN int
pam_sm_setcred(pam_handle_t *pamh, int flags,
int argc, const char *argv[])
{
      (void)pamh;
      (void)flags;
      (void)argc;
      (void)argv;
      return (PAM_SUCCESS);
}
