/*
*  Copyright 2016, University of Texas at Austin, W. Cyrus Proctor
*  Author: W. Cyrus Proctor <cproctor@tacc.utexas.edu>
*
*    This file is part of OpenMFA.
* 
*    OpenMFA is free software: you can redistribute it and/or
*    modify it under the terms of the GNU Affero General Public
*    License as published by the Free Software Foundation, either
*    version 3 of the License, or (at your option) any later
*    version.
* 
*    OpenMFA is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty
*    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
*    the GNU Affero General Public License for more details.
* 
*    You should have received a copy of the GNU Affero General
*    Public License along with OpenMFA. If not, see
*    <http://www.gnu.org/licenses/>.
*/

/*-
* Copyright 1998 Juniper Networks, Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
*
* THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
* FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
* OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
* LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
* OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
* SUCH DAMAGE.
*/


#include "radius.h"
#ifndef __USE_BSD
#define __USE_BSD
#endif
#include <sys/time.h>

//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------

static void
generr(struct rad_handle *h, const char *format, ...)
{
	va_list		 ap;
	va_start(ap, format);
	vsnprintf(h->errmsg, ERRSIZE, format, ap);
  openmfa_log(NULL, OPENMFA_ERR, h->errmsg);
	va_end(ap);
}

//-----------------------------------------------------------------------------------

const char *
rad_strerror(struct rad_handle *h)
{
	return h->errmsg;
}

//-----------------------------------------------------------------------------------

static void
clear_password(struct rad_handle *h)
{
	if (h->pass_len != 0) {
		memset(h->pass, 0, h->pass_len);
		h->pass_len = 0;
	}
	h->pass_pos = 0;
}

//-----------------------------------------------------------------------------------

char *
rad_cvt_string(const void *data, size_t len)
{
	char *s;

	s = malloc(len + 1);
	if (s != NULL) {
		memcpy(s, data, len);
		s[len] = '\0';
	}
	return s;
}

//-----------------------------------------------------------------------------------

struct rad_handle *
rad_auth_open(void)
{
	struct rad_handle *h;
	struct timeval tv;

	h = (struct rad_handle *)malloc(sizeof(struct rad_handle));
	if (h != NULL) {
		gettimeofday(&tv, NULL);
		srandom(tv.tv_sec ^ tv.tv_usec);
		h->fd = -1;
		h->num_servers = 0;
		h->ident = random();
		h->errmsg[0] = '\0';
		memset(h->pass, 0, sizeof h->pass);
		h->pass_len = 0;
		h->pass_pos = 0;
		h->chap_pass = 0;
		h->authentic_pos = 0;
		h->type = RADIUS_AUTH;
		h->out_created = 0;
		h->eap_msg = 0;
		h->bindto = INADDR_ANY;
	}
	return h;
}

//-----------------------------------------------------------------------------------

int
rad_add_server(struct rad_handle *h, const char *host, int port,
    const char *secret, int timeout, int tries)
{
    	struct in_addr bindto;
	bindto.s_addr = INADDR_ANY;

	return rad_add_server_ex(h, host, port, secret, timeout, tries,
		DEAD_TIME, &bindto);
}

//-----------------------------------------------------------------------------------

int
rad_add_server_ex(struct rad_handle *h, const char *host, int port,
    const char *secret, int timeout, int tries, int dead_time,
    struct in_addr *bindto)
{
	struct rad_server *srvp;

	if (h->num_servers >= MAXSERVERS) {
		generr(h, "Too many RADIUS servers specified");
		return -1;
	}
	srvp = &h->servers[h->num_servers];

	memset(&srvp->addr, 0, sizeof srvp->addr);
	srvp->addr.sin_family = AF_INET;
	if (!inet_aton(host, &srvp->addr.sin_addr)) {
		struct hostent *hent;

		if ((hent = gethostbyname(host)) == NULL) {
			generr(h, "%s: host not found", host);
			return -1;
		}
		memcpy(&srvp->addr.sin_addr, hent->h_addr,
		    sizeof srvp->addr.sin_addr);
	}
	if (port != 0)
		srvp->addr.sin_port = htons((u_short)port);
	else {
		struct servent *sent;

		if (h->type == RADIUS_AUTH)
			srvp->addr.sin_port =
			    (sent = getservbyname("radius", "udp")) != NULL ?
				sent->s_port : htons(RADIUS_PORT);
		else
			srvp->addr.sin_port =
			    (sent = getservbyname("radacct", "udp")) != NULL ?
				sent->s_port : htons(RADACCT_PORT);
	}
	if ((srvp->secret = strdup(secret)) == NULL) {
		generr(h, "Out of memory");
		return -1;
	}
	srvp->timeout = timeout;
	srvp->max_tries = tries;
	srvp->num_tries = 0;
	srvp->is_dead = 0;
	srvp->dead_time = dead_time;
	srvp->next_probe = 0;
	srvp->bindto = bindto->s_addr;
	h->num_servers++;
	return 0;
}

//-----------------------------------------------------------------------------------

int
rad_create_request(struct rad_handle *h, int code)
{
	int i;

	if (h->type == RADIUS_SERVER) {
		generr(h, "denied function call");
		return (-1);
	}
	if (h->num_servers == 0) {
	    	generr(h, "No RADIUS servers specified");
		return (-1);
	}
	h->out[POS_CODE] = code;
	h->out[POS_IDENT] = ++h->ident;
	if (code == RAD_ACCESS_REQUEST) {
		/* Create a random authenticator */
		for (i = 0;  i < LEN_AUTH;  i += 2) {
			long r;
			r = random();
			h->out[POS_AUTH+i] = (u_char)r;
			h->out[POS_AUTH+i+1] = (u_char)(r >> 8);
		}
	} else
		memset(&h->out[POS_AUTH], 0, LEN_AUTH);
	h->out_len = POS_ATTRS;
	clear_password(h);
	h->authentic_pos = 0;
	h->out_created = 1;
	return 0;
}

//-----------------------------------------------------------------------------------

static int
put_raw_attr(struct rad_handle *h, int type, const void *value, size_t len)
{
	if (len > 253) {
		generr(h, "Attribute too long");
		return -1;
	}
	if (h->out_len + 2 + len > MSGSIZE) {
		generr(h, "Maximum message length exceeded");
		return -1;
	}
	h->out[h->out_len++] = type;
	h->out[h->out_len++] = len + 2;
	memcpy(&h->out[h->out_len], value, len);
	h->out_len += len;
	return 0;
}

//-----------------------------------------------------------------------------------

static int
put_password_attr(struct rad_handle *h, int type, const void *value, size_t len)
{
	int padded_len;
	int pad_len;

	if (h->pass_pos != 0) {
		generr(h, "Multiple User-Password attributes specified");
		return -1;
	}
	if (len > PASSSIZE)
		len = PASSSIZE;
	padded_len = len == 0 ? 16 : (len+15) & ~0xf;
	pad_len = padded_len - len;

	/*
	 * Put in a place-holder attribute containing all zeros, and
	 * remember where it is so we can fill it in later.
	 */
	clear_password(h);
	put_raw_attr(h, type, h->pass, padded_len);
	h->pass_pos = h->out_len - padded_len;

	/* Save the cleartext password, padded as necessary */
	memcpy(h->pass, value, len);
	h->pass_len = len;
	memset(h->pass + len, 0, pad_len);
	return 0;
}

//-----------------------------------------------------------------------------------

int
rad_put_message_authentic(struct rad_handle *h)
{
#ifdef WITH_SSL
	u_char md_zero[MD5_DIGEST_LENGTH];

	if (h->out[POS_CODE] == RAD_ACCOUNTING_REQUEST) {
		generr(h, "Message-Authenticator is not valid"
		    " in accounting requests");
		return -1;
	}

	if (h->authentic_pos == 0) {
		h->authentic_pos = h->out_len;
		memset(md_zero, 0, sizeof(md_zero));
		return (put_raw_attr(h, RAD_MESSAGE_AUTHENTIC, md_zero,
		    sizeof(md_zero)));
	}
	return 0;
#else
	generr(h, "Message Authenticator not supported,"
	    " please recompile libradius with SSL support");
	return -1;
#endif
}

//-----------------------------------------------------------------------------------

/*
 * Returns the attribute type.  If none are left, returns 0.  On failure,
 * returns -1.
 */
int
rad_get_attr(struct rad_handle *h, const void **value, size_t *len)
{
	int type;

	if (h->in_pos >= h->in_len)
		return 0;
	if (h->in_pos + 2 > h->in_len) {
		generr(h, "Malformed attribute in response");
		return -1;
	}
	type = h->in[h->in_pos++];
	*len = h->in[h->in_pos++] - 2;
	if (h->in_pos + (int)*len > h->in_len) {
		generr(h, "Malformed attribute in response");
		return -1;
	}
	*value = &h->in[h->in_pos];
	h->in_pos += *len;
	return type;
}

//-----------------------------------------------------------------------------------

int
rad_put_attr(struct rad_handle *h, int type, const void *value, size_t len)
{
	int result;

	if (!h->out_created) {
		generr(h, "Please call rad_create_request()"
		    " before putting attributes");
		return -1;
	}

	if (h->out[POS_CODE] == RAD_ACCOUNTING_REQUEST) {
		if (type == RAD_EAP_MESSAGE) {
			generr(h, "EAP-Message attribute is not valid"
			    " in accounting requests");
			return -1;
		}
	}

	/*
	 * When proxying EAP Messages, the Message Authenticator
	 * MUST be present; see RFC 3579.
	 */
	if (type == RAD_EAP_MESSAGE) {
		if (rad_put_message_authentic(h) == -1)
			return -1;
	}

	if (type == RAD_USER_PASSWORD) {
		result = put_password_attr(h, type, value, len);
	} else if (type == RAD_MESSAGE_AUTHENTIC) {
		result = rad_put_message_authentic(h);
	} else {
		result = put_raw_attr(h, type, value, len);
		if (result == 0) {
			if (type == RAD_CHAP_PASSWORD)
				h->chap_pass = 1;
			else if (type == RAD_EAP_MESSAGE)
				h->eap_msg = 1;
		}
	}

	return result;
}

//-----------------------------------------------------------------------------------

int
rad_put_string(struct rad_handle *h, int type, const char *str)
{
	return rad_put_attr(h, type, str, strlen(str));
}

//-----------------------------------------------------------------------------------

static void
insert_request_authenticator(struct rad_handle *h, int resp)
{
	MD5_CTX ctx;
	const struct rad_server *srvp;

	srvp = &h->servers[h->srv];

	/* Create the request authenticator */
	MD5_Init(&ctx);
	MD5_Update(&ctx, &h->out[POS_CODE], POS_AUTH - POS_CODE);
	if (resp)
	    MD5_Update(&ctx, &h->in[POS_AUTH], LEN_AUTH);
	else
	    MD5_Update(&ctx, &h->out[POS_AUTH], LEN_AUTH);
	MD5_Update(&ctx, &h->out[POS_ATTRS], h->out_len - POS_ATTRS);
	MD5_Update(&ctx, srvp->secret, strlen(srvp->secret));
	MD5_Final(&h->out[POS_AUTH], &ctx);
}

//-----------------------------------------------------------------------------------

static void
insert_message_authenticator(struct rad_handle *h, int resp)
{
#ifdef WITH_SSL
	u_char md[EVP_MAX_MD_SIZE];
	u_int md_len;
	const struct rad_server *srvp;
	HMAC_CTX *ctx;
	srvp = &h->servers[h->srv];

	if (h->authentic_pos != 0) {
		ctx = HMAC_CTX_new();
		HMAC_Init_ex(ctx, srvp->secret, strlen(srvp->secret), EVP_md5(), NULL);
		HMAC_Update(ctx, &h->out[POS_CODE], POS_AUTH - POS_CODE);
		if (resp)
		    HMAC_Update(ctx, &h->in[POS_AUTH], LEN_AUTH);
		else
		    HMAC_Update(ctx, &h->out[POS_AUTH], LEN_AUTH);
		HMAC_Update(ctx, &h->out[POS_ATTRS],
		    h->out_len - POS_ATTRS);
		HMAC_Final(ctx, md, &md_len);
		HMAC_CTX_free(ctx);
		memcpy(&h->out[h->authentic_pos + 2], md, md_len);
	}
#endif
}

//-----------------------------------------------------------------------------------

static void
insert_scrambled_password(struct rad_handle *h, int srv)
{
	MD5_CTX ctx;
	unsigned char md5[MD5_DIGEST_LENGTH];
	const struct rad_server *srvp;
	int padded_len;
	int pos;

	srvp = &h->servers[srv];
	padded_len = h->pass_len == 0 ? 16 : (h->pass_len+15) & ~0xf;

	memcpy(md5, &h->out[POS_AUTH], LEN_AUTH);
	for (pos = 0;  pos < padded_len;  pos += 16) {
		int i;

		/* Calculate the new scrambler */
		MD5_Init(&ctx);
		MD5_Update(&ctx, srvp->secret, strlen(srvp->secret));
		MD5_Update(&ctx, md5, 16);
		MD5_Final(md5, &ctx);

		/*
		 * Mix in the current chunk of the password, and copy
		 * the result into the right place in the request.  Also
		 * modify the scrambler in place, since we will use this
		 * in calculating the scrambler for next time.
		 */
		for (i = 0;  i < 16;  i++)
			h->out[h->pass_pos + pos + i] =
			    md5[i] ^= h->pass[pos + i];
	}
}

//-----------------------------------------------------------------------------------

/*
 * Return true if the current response is valid for a request to the
 * specified server.
 */
static int
is_valid_response(struct rad_handle *h, int srv,
    const struct sockaddr_in *from)
{
	MD5_CTX ctx;
	unsigned char md5[MD5_DIGEST_LENGTH];
	const struct rad_server *srvp;
	int len;
#ifdef WITH_SSL
	HMAC_CTX *hctx;
	u_char resp[MSGSIZE], md[EVP_MAX_MD_SIZE];
	u_int md_len;
	int pos;
#endif

	srvp = &h->servers[srv];

	/* Check the source address */
	if (from->sin_family != srvp->addr.sin_family ||
	    from->sin_addr.s_addr != srvp->addr.sin_addr.s_addr ||
	    from->sin_port != srvp->addr.sin_port)
		return 0;

	/* Check the message length */
	if (h->in_len < POS_ATTRS)
		return 0;
	len = h->in[POS_LENGTH] << 8 | h->in[POS_LENGTH+1];
	if (len > h->in_len)
		return 0;

	/* Check the response authenticator */
	MD5_Init(&ctx);
	MD5_Update(&ctx, &h->in[POS_CODE], POS_AUTH - POS_CODE);
	MD5_Update(&ctx, &h->out[POS_AUTH], LEN_AUTH);
	MD5_Update(&ctx, &h->in[POS_ATTRS], len - POS_ATTRS);
	MD5_Update(&ctx, srvp->secret, strlen(srvp->secret));
	MD5_Final(md5, &ctx);
	if (memcmp(&h->in[POS_AUTH], md5, sizeof md5) != 0)
		return 0;

#ifdef WITH_SSL
	/*
	 * For non accounting responses check the message authenticator,
	 * if any.
	 */
	if (h->in[POS_CODE] != RAD_ACCOUNTING_RESPONSE) {

		memcpy(resp, h->in, MSGSIZE);
		pos = POS_ATTRS;

		/* Search and verify the Message-Authenticator */
		while (pos < len - 2) {

			if (h->in[pos] == RAD_MESSAGE_AUTHENTIC) {
				/* zero fill the Message-Authenticator */
				memset(&resp[pos + 2], 0, MD5_DIGEST_LENGTH);

				hctx = HMAC_CTX_new();
				HMAC_Init_ex(hctx, srvp->secret,
				    strlen(srvp->secret), EVP_md5(), NULL);
				HMAC_Update(hctx, &h->in[POS_CODE],
				    POS_AUTH - POS_CODE);
				HMAC_Update(hctx, &h->out[POS_AUTH],
				    LEN_AUTH);
				HMAC_Update(hctx, &resp[POS_ATTRS],
				    h->in_len - POS_ATTRS);
				HMAC_Final(hctx, md, &md_len);
				HMAC_CTX_free(hctx);
				if (memcmp(md, &h->in[pos + 2],
				    MD5_DIGEST_LENGTH) != 0)
					return 0;
				break;
			}
			pos += h->in[pos + 1];
		}
	}
#endif
	return 1;
}

//-----------------------------------------------------------------------------------

/*
 * rad_init_send_request() must have previously been called.
 * Returns:
 *   0     The application should select on *fd with a timeout of tv before
 *         calling rad_continue_send_request again.
 *   < 0   Failure
 *   > 0   Success
 */
int
rad_continue_send_request(struct rad_handle *h, int selected, int *fd,
                          struct timeval *tv)
{
	int n, cur_srv;
	time_t now;
	struct sockaddr_in sin;

	if (h->type == RADIUS_SERVER) {
		generr(h, "denied function call");
		return (-1);
	}
	if (selected) {
		struct sockaddr_in from;
		socklen_t fromlen;

		fromlen = sizeof from;
		h->in_len = recvfrom(h->fd, h->in,
		    MSGSIZE, MSG_WAITALL, (struct sockaddr *)&from, &fromlen);
		if (h->in_len == -1) {
			generr(h, "recvfrom: %s", strerror(errno));
			return -1;
		}
		if (is_valid_response(h, h->srv, &from)) {
			h->in_len = h->in[POS_LENGTH] << 8 |
			    h->in[POS_LENGTH+1];
			h->in_pos = POS_ATTRS;
			return h->in[POS_CODE];
		}
	}

	/*
         * Scan round-robin to the next server that has some
         * tries left.  There is guaranteed to be one, or we
         * would have exited this loop by now.
	 */
	cur_srv = h->srv;
	now = time(NULL);
	if (h->servers[h->srv].num_tries >= h->servers[h->srv].max_tries) {
		/* Set next probe time for this server */
		if (h->servers[h->srv].dead_time) {
			h->servers[h->srv].is_dead = 1;
			h->servers[h->srv].next_probe = now +
			    h->servers[h->srv].dead_time;
		}
		do {
		    	h->srv++;
			if (h->srv >= h->num_servers)
				h->srv = 0;
			if (h->servers[h->srv].is_dead == 0)
			    	break;
			if (h->servers[h->srv].dead_time &&
			    h->servers[h->srv].next_probe <= now) {
			    	h->servers[h->srv].is_dead = 0;
				h->servers[h->srv].num_tries = 0;
				break;
			}
		} while (h->srv != cur_srv);

		if (h->srv == cur_srv) {
			generr(h, "No valid RADIUS responses received");
			return (-1);
		}
	}

	/* Rebind */
	if (h->bindto != h->servers[h->srv].bindto) {
	    	h->bindto = h->servers[h->srv].bindto;
		close(h->fd);
		if ((h->fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1) {
			generr(h, "Cannot create socket: %s", strerror(errno));
			return -1;
		}
		memset(&sin, 0, sizeof sin);
		sin.sin_family = AF_INET;
		sin.sin_addr.s_addr = h->bindto;
		sin.sin_port = 0;
		if (bind(h->fd, (const struct sockaddr *)&sin,
		    sizeof sin) == -1) {
			generr(h, "bind: %s", strerror(errno));
			close(h->fd);
			h->fd = -1;
			return (-1);
		}
	}

	if (h->out[POS_CODE] == RAD_ACCESS_REQUEST) {
		/* Insert the scrambled password into the request */
		if (h->pass_pos != 0)
			insert_scrambled_password(h, h->srv);
	}
	insert_message_authenticator(h, 0);

	if (h->out[POS_CODE] != RAD_ACCESS_REQUEST) {
		/* Insert the request authenticator into the request */
		memset(&h->out[POS_AUTH], 0, LEN_AUTH);
		insert_request_authenticator(h, 0);
	}

	/* Send the request */
	n = sendto(h->fd, h->out, h->out_len, 0,
	    (const struct sockaddr *)&h->servers[h->srv].addr,
	    sizeof h->servers[h->srv].addr);
	if (n != h->out_len)
		tv->tv_sec = 1; /* Do not wait full timeout if send failed. */
	else
		tv->tv_sec = h->servers[h->srv].timeout;
	h->servers[h->srv].num_tries++;
	tv->tv_usec = 0;
	*fd = h->fd;

	return 0;
}

//-----------------------------------------------------------------------------------

/*
 * Returns -1 on error, 0 to indicate no event and >0 for success
 */
int
rad_init_send_request(struct rad_handle *h, int *fd, struct timeval *tv)
{
	int srv;
	time_t now;
	struct sockaddr_in sin;

	if (h->type == RADIUS_SERVER) {
		generr(h, "denied function call");
		return (-1);
	}
	/* Make sure we have a socket to use */
	if (h->fd == -1) {
		if ((h->fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1) {
			generr(h, "Cannot create socket: %s", strerror(errno));
			return -1;
		}
		memset(&sin, 0, sizeof sin);
		sin.sin_family = AF_INET;
		sin.sin_addr.s_addr = h->bindto;
		sin.sin_port = htons(0);
		if (bind(h->fd, (const struct sockaddr *)&sin,
		    sizeof sin) == -1) {
			generr(h, "bind: %s", strerror(errno));
			close(h->fd);
			h->fd = -1;
			return -1;
		}
	}

	if (h->out[POS_CODE] != RAD_ACCESS_REQUEST) {
		/* Make sure no password given */
		if (h->pass_pos || h->chap_pass) {
			generr(h, "User or Chap Password"
			    " in accounting request");
			return -1;
		}
	} else {
		if (h->eap_msg == 0) {
			/* Make sure the user gave us a password */
			if (h->pass_pos == 0 && !h->chap_pass) {
				generr(h, "No User or Chap Password"
				    " attributes given");
				return -1;
			}
			if (h->pass_pos != 0 && h->chap_pass) {
				generr(h, "Both User and Chap Password"
				    " attributes given");
				return -1;
			}
		}
	}

	/* Fill in the length field in the message */
	h->out[POS_LENGTH] = h->out_len >> 8;
	h->out[POS_LENGTH+1] = h->out_len;

	h->srv = 0;
	now = time(NULL);
	for (srv = 0;  srv < h->num_servers;  srv++)
		h->servers[srv].num_tries = 0;
	/* Find a first good server. */
	for (srv = 0;  srv < h->num_servers;  srv++) {
		if (h->servers[srv].is_dead == 0)
			break;
		if (h->servers[srv].dead_time &&
		    h->servers[srv].next_probe <= now) {
		    	h->servers[srv].is_dead = 0;
			break;
		}
		h->srv++;
	}

	/* If all servers was dead on the last probe, try from beginning */
	if (h->srv == h->num_servers) {
		for (srv = 0;  srv < h->num_servers;  srv++) {
		    	h->servers[srv].is_dead = 0;
			h->servers[srv].next_probe = 0;
		}
		h->srv = 0;
	}

	return rad_continue_send_request(h, 0, fd, tv);
}

//-----------------------------------------------------------------------------------

/*
 * Returns the response type code on success, or -1 on failure.
 */
int
rad_send_request(struct rad_handle *h)
{
	struct timeval timelimit;
	struct timeval tv;
	int fd;
	int n;

	n = rad_init_send_request(h, &fd, &tv);

	if (n != 0)
		return n;

	gettimeofday(&timelimit, NULL);
	timeradd(&tv, &timelimit, &timelimit);

	for ( ; ; ) {
		fd_set readfds;

		FD_ZERO(&readfds);
		FD_SET(fd, &readfds);

		n = select(fd + 1, &readfds, NULL, NULL, &tv);

		if (n == -1) {
			generr(h, "select: %s", strerror(errno));
			return -1;
		}

		if (!FD_ISSET(fd, &readfds)) {
			/* Compute a new timeout */
			gettimeofday(&tv, NULL);
			timersub(&timelimit, &tv, &tv);
			if (tv.tv_sec > 0 || (tv.tv_sec == 0 && tv.tv_usec > 0))
				/* Continue the select */
				continue;
		}

		n = rad_continue_send_request(h, n, &fd, &tv);

		if (n != 0)
			return n;

		gettimeofday(&timelimit, NULL);
		timeradd(&tv, &timelimit, &timelimit);
	}
}

