/*
*  Copyright 2016, University of Texas at Austin, W. Cyrus Proctor
*  Author: W. Cyrus Proctor <cproctor@tacc.utexas.edu>
*
*    This file is part of OpenMFA.
* 
*    OpenMFA is free software: you can redistribute it and/or
*    modify it under the terms of the GNU Affero General Public
*    License as published by the Free Software Foundation, either
*    version 3 of the License, or (at your option) any later
*    version.
* 
*    OpenMFA is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty
*    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
*    the GNU Affero General Public License for more details.
* 
*    You should have received a copy of the GNU Affero General
*    Public License along with OpenMFA. If not, see
*    <http://www.gnu.org/licenses/>.
*/

/*-
* Copyright 1998 Juniper Networks, Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
*
* THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
* FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
* OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
* LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
* OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
* SUCH DAMAGE.
*/


#ifndef _RADIUS_H_
#define _RADIUS_H_

#include <errno.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <unistd.h>
#ifdef WITH_SSL
#include <openssl/ossl_typ.h>
#include <openssl/hmac.h>
#include <openssl/md5.h>
#include <openssl/opensslv.h>
#else
#define MD5_DIGEST_LENGTH 16
#include "md5/md5.h"
#endif

#include "pam_openmfa.h"

//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------

// Handle types 
#define RADIUS_AUTH   0 // RADIUS authentication, default 
#define RADIUS_ACCT   1 // RADIUS accounting 
#define RADIUS_SERVER 2 // RADIUS server 

//-----------------------------------------------------------------------------------

// Limits 
#define ERRSIZE     128  // Maximum error message length 
#define MAXCONFLINE 1024 // Maximum config file line length 
#define MAXSERVERS  10   // Maximum number of servers to try 
#define MSGSIZE     4096 // Maximum RADIUS message 
#define PASSSIZE    128  // Maximum significant password chars 

//-----------------------------------------------------------------------------------

// Message types 
#define RAD_ACCESS_REQUEST      1
#define RAD_ACCESS_ACCEPT       2
#define RAD_ACCESS_REJECT       3
#define RAD_ACCOUNTING_REQUEST  4
#define RAD_ACCOUNTING_RESPONSE 5
#define RAD_ACCESS_CHALLENGE    11
#define RAD_DISCONNECT_REQUEST  40
#define RAD_DISCONNECT_ACK      41
#define RAD_DISCONNECT_NAK      42
#define RAD_COA_REQUEST         43
#define RAD_COA_ACK             44
#define RAD_COA_NAK             45

//-----------------------------------------------------------------------------------

// Attribute types and values 
#define RAD_USER_NAME         1   // String 
#define RAD_USER_PASSWORD     2   // String 
#define RAD_CHAP_PASSWORD     3   // String 
#define RAD_NAS_IP_ADDRESS    4   // IP address 
#define RAD_NAS_PORT          5   // Integer 
#define RAD_SERVICE_TYPE      6   // Integer 
#define RAD_EAP_MESSAGE       79  // Octets 
#define RAD_MESSAGE_AUTHENTIC 80  // Octets 

//-----------------------------------------------------------------------------------

// Defaults 
#define MAXTRIES         3
#define PATH_RADIUS_CONF "/etc/radius.conf"
#define RADIUS_PORT      1812
#define RADACCT_PORT     1813
#define TIMEOUT          3 // In seconds 
#define  DEAD_TIME       0

//-----------------------------------------------------------------------------------

// Positions of fields in RADIUS messages 
#define POS_CODE   0    // Message code 
#define POS_IDENT  1    // Identifier 
#define POS_LENGTH 2    // Message length 
#define POS_AUTH   4    // Authenticator 
#define LEN_AUTH   16   // Length of authenticator 
#define POS_ATTRS  20   // Start of attributes 

//-----------------------------------------------------------------------------------

struct rad_server {
  struct     sockaddr_in addr;  // Address of server 
  char      *secret;            // Shared secret 
  int        timeout;           // Timeout in seconds 
  int        max_tries;         // Number of tries before giving up 
  int        num_tries;         // Number of tries so far 
  int        is_dead;           // The server did not answer last time 
  time_t     dead_time;         // Don't try this server for the time period if it is dead 
  time_t     next_probe;        // Time of a next probe after failure 
  in_addr_t  bindto;            // Bind to address 
};

//-----------------------------------------------------------------------------------

struct rad_handle {
  int           fd;                      // Socket file descriptor 
  struct rad_server servers[MAXSERVERS]; // Servers to contact 
  int           num_servers;             // Number of valid server entries 
  int           ident;                   // Current identifier value 
  char          errmsg[ERRSIZE];         // Most recent error message 
  unsigned char out[MSGSIZE];            // Request to send 
  char          out_created;             // rad_create_request() called? 
  int           out_len;                 // Length of request 
  char          pass[PASSSIZE];          // Cleartext password 
  int           pass_len;                // Length of cleartext password 
  int           pass_pos;                // Position of scrambled password 
  char          chap_pass;               // Have we got a CHAP_PASSWORD ? 
  int           authentic_pos;           // Position of message authenticator 
  char          eap_msg;                 // Are we an EAP Proxy? 
  unsigned char in[MSGSIZE];             // Response received 
  int           in_len;                  // Length of response 
  int           in_pos;                  // Current position scanning attrs 
  int           srv;                     // Server number we did last 
  int           type;                    // Handle type 
  in_addr_t    bindto;                   // Current bind address 
};


//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------


char              *rad_cvt_string               (const void *, size_t);
const char        *rad_strerror                 (struct rad_handle *);
int                rad_add_server               (struct rad_handle *, const char *, int, const char *, int, int);
int                rad_add_server_ex            (struct rad_handle *, const char *, int, const char *, int, int, int, struct in_addr *);
int                rad_continue_send_request    (struct rad_handle *, int, int *, struct timeval *);
int                rad_create_request           (struct rad_handle *, int);
int                rad_get_attr                 (struct rad_handle *, const void **, size_t *);
int                rad_init_send_request        (struct rad_handle *, int *, struct timeval *);
int                rad_put_attr                 (struct rad_handle *, int, const void *, size_t);
int                rad_put_message_authentic    (struct rad_handle *);
int                rad_put_string               (struct rad_handle *, int, const char *);
int                rad_send_request             (struct rad_handle *);
static int         is_valid_response            (struct rad_handle *, int, const struct sockaddr_in *);
static int         put_password_attr            (struct rad_handle *, int, const void *, size_t);
static int         put_raw_attr                 (struct rad_handle *, int, const void *, size_t);
static void        clear_password               (struct rad_handle *);
static void        generr                       (struct rad_handle *, const char *, ...);
static void        insert_message_authenticator (struct rad_handle *, int);
static void        insert_request_authenticator (struct rad_handle *, int);
static void        insert_scrambled_password    (struct rad_handle *, int);
struct rad_handle *rad_auth_open                (void);

//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------

#endif // _RADIUS_H_
